==> Pcore-DBH-v0.1.1.tar.gz: Pcore-DBH-v0.1.1/lib/Pcore/DBH/DDL.pm <==
package Pcore::DBH::DDL;

use Pcore -role;
use Pcore::DBH::DDL::ChangeSet;

requires qw[schema_info_sql];

has dbh => ( is => 'ro', isa => ConsumerOf ['Pcore::DBH'], required => 1 );

has _schema_info_table => ( is => 'lazy', isa => Str, default => '_schema_info', init_arg => undef );
has _changesets => ( is => 'lazy', isa => HashRef, default => sub { {} }, init_arg => undef );

# TODO
# revert all changes if one changeset failed
# upgrade db to specified version, including downgrade - need to specify revert sql code

sub add_changeset ( $self, %args ) {
    my $cset = Pcore::DBH::DDL::ChangeSet->new( \%args );

    my $id = $cset->component . $cset->id;

    die q[DDL changeset id "] . $cset->id . q[" for component "] . $cset->component . q[" already exists] if exists $self->_changesets->{$id};

    $self->_changesets->{$id} = $cset;

    return $cset;
}

sub upgrade ($self) {
    $self->_create_schema_info;

    my $dbh = $self->dbh->dbh;

    my $info = $dbh->query( 'SELECT * FROM', [ $self->_schema_info_table ] )->selectall_hashref( key_cols => 'component' ) // {};

    for my $cset ( sort { $a->id <=> $b->id } values $self->_changesets->%* ) {
        $dbh->query( 'INSERT INTO', [ $self->_schema_info_table ], VALUES => { component => $cset->component } )->do if !exists $info->{ $cset->component };

        if ( !exists $info->{ $cset->component } || !defined $info->{ $cset->component }->{changeset} || $info->{ $cset->component }->{changeset} < $cset->id ) {
            $dbh->begin_work if $cset->transaction;

            my $error = try {
                if ( ref $cset->sql eq 'CODE' ) {
                    die q[Changeset "] . $cset->id . q[" for component "] . $cset->component . q[" did not return true value] if !$cset->sql->( $cset, $dbh );
                }
                else {
                    $self->_get_changeset_query($cset)->do( cache => 0 );
                }

                $dbh->commit if $cset->transaction;

                return 0;
            }
            catch {
                my $e = shift;

                $e->sendlog;

                $dbh->rollback if $cset->transaction;

                return 1;
            };

            die q[Failed to apply changeset "] . $cset->id . q[" for component "] . $cset->component . q["] if $error;

            # update schema info
            $dbh->query( UPDATE => [ $self->_schema_info_table ], SET => { changeset => $cset->id }, 'WHERE component =', \$cset->component )->do;

            $info->{ $cset->component }->{changeset} = $cset->id;
        }
    }

    return;
}

sub _create_schema_info ($self) {
    return $self->dbh->query( $self->schema_info_sql->$* )->do;
}

sub _get_changeset_query ( $self, $cset ) {
    return $self->dbh->query( $self->_get_cset_sql($cset) );
}

sub _get_cset_sql ( $self, $cset ) {
    return $cset->sql;
}

1;
## -----SOURCE FILTER LOG BEGIN-----
##
## PerlCritic profile "pcore-script" policy violations:
## +------+----------------------+----------------------------------------------------------------------------------------------------------------+
## | Sev. | Lines                | Policy                                                                                                         |
## |======+======================+================================================================================================================|
## |    3 | 36                   | References::ProhibitDoubleSigils - Double-sigil dereference                                                    |
## +------+----------------------+----------------------------------------------------------------------------------------------------------------+
##
## -----SOURCE FILTER LOG END-----
__END__
=pod

=encoding utf8

=head1 NAME

Pcore::DBH::DDL - database schema versioning subsystem

=head1 SYNOPSIS

    my $ddl = $dbh->ddl;

    $ddl->add_changeset(
        id        => 1,
        component => undef,
        sql       => q[CREATE TABLE ...],
    );

    ...

    $ddl->upgrade;

=head1 DESCRIPTION

=cut

==> Pcore-DBH-v0.1.2.tar.gz: Pcore-DBH-v0.1.2/lib/Pcore/DBH/DDL.pm <==
package Pcore::DBH::DDL;

use Pcore -role;
use Pcore::DBH::DDL::ChangeSet;

requires qw[schema_info_sql];

has dbh => ( is => 'ro', isa => ConsumerOf ['Pcore::DBH'], required => 1 );

has dbc => ( is => 'lazy', isa => ConsumerOf ['Pcore::DBH'], init_arg => undef );    # effective dbh
has _schema_info_table => ( is => 'lazy', isa => Str, default => '_schema_info', init_arg => undef );
has _changesets => ( is => 'lazy', isa => HashRef, default => sub { {} }, init_arg => undef );

# TODO
# revert all changes if one changeset failed
# upgrade db to specified version, including downgrade - need to specify revert sql code

sub _build_dbc($self) {
    return $self->dbh->dbh;
}

sub add_changeset ( $self, %args ) {
    my $cset = Pcore::DBH::DDL::ChangeSet->new( \%args );

    my $id = $cset->component . $cset->id;

    die qq[DDL changeset id "@{[$cset->id]}" for component "@{[$cset->component]}" already exists] if exists $self->_changesets->{$id};

    $self->_changesets->{$id} = $cset;

    return $cset;
}

sub upgrade ($self) {
    $self->_create_schema_info;

    my $dbh = $self->dbc;

    my $info = $dbh->selectall_hashref( [ 'SELECT * FROM', [ $self->_schema_info_table ] ], key_cols => 'component' ) // {};

    for my $cset ( sort { $a->id <=> $b->id } values $self->_changesets->%* ) {

        # create component record
        $dbh->do( [ 'INSERT INTO', [ $self->_schema_info_table ], VALUES => { component => $cset->component } ] ) if !exists $info->{ $cset->component };

        if ( !exists $info->{ $cset->component } || !defined $info->{ $cset->component }->{changeset} || $info->{ $cset->component }->{changeset} < $cset->id ) {
            $dbh->begin_work if $cset->transaction;

            my $error = try {
                if ( ref $cset->sql eq 'CODE' ) {
                    die qq[Changeset "@{[$cset->id]}" for component "@{[$cset->component]}" did not return true value] if !$cset->sql->( $cset, $dbh );
                }
                else {
                    $dbh->do( $self->_get_changeset_query($cset), cache => 0 );
                }

                $dbh->commit if $cset->transaction;

                return 0;
            }
            catch {
                my $e = shift;

                $e->sendlog;

                $dbh->rollback if $cset->transaction;

                return 1;
            };

            die qq[Failed to apply changeset "@{[$cset->id]}" for component "@{[$cset->component]}"] if $error;

            # update schema info
            $dbh->do( [ UPDATE => [ $self->_schema_info_table ], SET => { changeset => $cset->id }, 'WHERE component =', \$cset->component ] );

            $info->{ $cset->component }->{changeset} = $cset->id;
        }
    }

    return;
}

sub _create_schema_info ($self) {
    return $self->dbh->do( $self->schema_info_sql->$*, cache => 0 );
}

sub _get_changeset_query ( $self, $cset ) {
    return $self->dbh->query( $self->_get_cset_sql($cset) );
}

sub _get_cset_sql ( $self, $cset ) {
    return $cset->sql;
}

1;
## -----SOURCE FILTER LOG BEGIN-----
##
## PerlCritic profile "pcore-script" policy violations:
## +------+----------------------+----------------------------------------------------------------------------------------------------------------+
## | Sev. | Lines                | Policy                                                                                                         |
## |======+======================+================================================================================================================|
## |    3 | 41                   | References::ProhibitDoubleSigils - Double-sigil dereference                                                    |
## +------+----------------------+----------------------------------------------------------------------------------------------------------------+
##
## -----SOURCE FILTER LOG END-----
__END__
=pod

=encoding utf8

=head1 NAME

Pcore::DBH::DDL - database schema versioning subsystem

=head1 SYNOPSIS

    my $ddl = $dbh->ddl;

    $ddl->add_changeset(
        id        => 1,
        component => undef,
        sql       => q[CREATE TABLE ...],
    );

    ...

    $ddl->upgrade;

=head1 DESCRIPTION

=cut
